@@ -1,4 +1,6 @@
 import asyncio
+import collections
+
 
 from psycopg2.extensions import TRANSACTION_STATUS_IDLE
 
@@ -10,22 +12,26 @@ from .log import logger
 def create_pool(dsn=None, *, minsize=10, maxsize=10,
                 loop=None, timeout=TIMEOUT,
                 enable_json=True, enable_hstore=True,
+                echo=False,
                 **kwargs):
     if loop is None:
         loop = asyncio.get_event_loop()
 
     pool = Pool(dsn, minsize, maxsize, loop, timeout,
                 enable_json=enable_json, enable_hstore=enable_hstore,
+                echo=echo,
                 **kwargs)
-    yield from pool._fill_free_pool(False)
+    if minsize > 0:
+        with (yield from pool._cond):
+            yield from pool._fill_free_pool(False)
     return pool
 
 
-class Pool:
+class Pool(asyncio.AbstractServer):
     """Connection pool"""
 
     def __init__(self, dsn, minsize, maxsize, loop, timeout, *,
-                 enable_json, enable_hstore, **kwargs):
+                 enable_json, enable_hstore, echo, **kwargs):
         if minsize < 0:
             raise ValueError("minsize should be zero or greater")
         if maxsize < minsize:
@@ -36,9 +42,19 @@ class Pool:
         self._timeout = timeout
         self._enable_json = enable_json
         self._enable_hstore = enable_hstore
+        self._echo = echo
         self._conn_kwargs = kwargs
-        self._free = asyncio.queues.Queue(maxsize, loop=self._loop)
+        self._acquiring = 0
+        self._free = collections.deque(maxlen=maxsize)
+        self._cond = asyncio.Condition(loop=loop)
         self._used = set()
+        self._terminated = set()
+        self._closing = False
+        self._closed = False
+
+    @property
+    def echo(self):
+        return self._echo
 
     @property
     def minsize(self):
@@ -46,15 +62,15 @@ class Pool:
 
     @property
     def maxsize(self):
-        return self._free.maxsize
+        return self._free.maxlen
 
     @property
     def size(self):
-        return self.freesize + len(self._used)
+        return self.freesize + len(self._used) + self._acquiring
 
     @property
     def freesize(self):
-        return self._free.qsize()
+        return len(self._free)
 
     @property
     def timeout(self):
@@ -63,44 +79,121 @@ class Pool:
     @asyncio.coroutine
     def clear(self):
         """Close all free connections in pool."""
-        while not self._free.empty():
-            conn = yield from self._free.get()
-            yield from conn.close()
+        with (yield from self._cond):
+            while self._free:
+                conn = self._free.popleft()
+                yield from conn.close()
+            self._cond.notify()
+
+    def close(self):
+        """Close pool.
+
+        Mark all pool connections to be closed on getting back to pool.
+        Closed pool doesn't allow to acquire new connections.
+        """
+        if self._closed:
+            return
+        self._closing = True
+
+    def terminate(self):
+        """Terminate pool.
+
+        Close pool with instantly closing all acquired connections also.
+        """
+
+        self.close()
+
+        for conn in list(self._used):
+            conn.close()
+            self._terminated.add(conn)
+
+        self._used.clear()
+
+    @asyncio.coroutine
+    def wait_closed(self):
+        """Wait for closing all pool's connections."""
+
+        if self._closed:
+            return
+        if not self._closing:
+            raise RuntimeError(".wait_closed() should be called "
+                               "after .close()")
+
+        while self._free:
+            conn = self._free.popleft()
+            conn.close()
+
+        with (yield from self._cond):
+            while self.size > self.freesize:
+                yield from self._cond.wait()
+
+        self._closed = True
 
     @asyncio.coroutine
     def acquire(self):
         """Acquire free connection from the pool."""
-        yield from self._fill_free_pool(True)
-        conn = yield from self._free.get()
-        assert not conn.closed, conn
-        assert conn not in self._used, (conn, self._used)
-        self._used.add(conn)
-        return conn
+        if self._closing:
+            raise RuntimeError("Cannot acquire connection after closing pool")
+        with (yield from self._cond):
+            while True:
+                yield from self._fill_free_pool(True)
+                if self._free:
+                    conn = self._free.popleft()
+                    assert not conn.closed, conn
+                    assert conn not in self._used, (conn, self._used)
+                    self._used.add(conn)
+                    return conn
+                else:
+                    yield from self._cond.wait()
 
     @asyncio.coroutine
     def _fill_free_pool(self, override_min):
         while self.size < self.minsize:
-            conn = yield from connect(
-                self._dsn, loop=self._loop, timeout=self._timeout,
-                enable_json=self._enable_json,
-                enable_hstore=self._enable_hstore,
-                **self._conn_kwargs)
-            yield from self._free.put(conn)
-        if not self._free.empty():
+            self._acquiring += 1
+            try:
+                conn = yield from connect(
+                    self._dsn, loop=self._loop, timeout=self._timeout,
+                    enable_json=self._enable_json,
+                    enable_hstore=self._enable_hstore,
+                    echo=self._echo,
+                    **self._conn_kwargs)
+                # raise exception if pool is closing
+                self._free.append(conn)
+                self._cond.notify()
+            finally:
+                self._acquiring -= 1
+        if self._free:
             return
+
         if override_min and self.size < self.maxsize:
-            conn = yield from connect(
-                self._dsn, loop=self._loop, timeout=self._timeout,
-                enable_json=self._enable_json,
-                enable_hstore=self._enable_hstore,
-                **self._conn_kwargs)
-            yield from self._free.put(conn)
+            self._acquiring += 1
+            try:
+                conn = yield from connect(
+                    self._dsn, loop=self._loop, timeout=self._timeout,
+                    enable_json=self._enable_json,
+                    enable_hstore=self._enable_hstore,
+                    echo=self._echo,
+                    **self._conn_kwargs)
+                # raise exception if pool is closing
+                self._free.append(conn)
+                self._cond.notify()
+            finally:
+                self._acquiring -= 1
+
+    @asyncio.coroutine
+    def _wakeup(self):
+        with (yield from self._cond):
+            self._cond.notify()
 
     def release(self, conn):
         """Release free connection back to the connection pool.
 
         This is NOT a coroutine.
         """
+        if conn in self._terminated:
+            assert conn.closed, conn
+            self._terminated.remove(conn)
+            return
         assert conn in self._used, (conn, self._used)
         self._used.remove(conn)
         if not conn.closed:
@@ -109,20 +202,13 @@ class Pool:
                 logger.warning(
                     "Invalid transaction status on released connection: %d",
                     tran_status)
-                conn._close()
+                conn.close()
                 return
-            while True:
-                try:
-                    self._free.put_nowait(conn)
-                except asyncio.QueueFull:
-                    # close the oldest connection in free pool.
-                    # that can happen if we acuire multiple
-                    # connections from parallel tasks
-                    # when minsize == 0
-                    conn2 = self._free.get_nowait()
-                    conn2._close()
-                else:
-                    break
+            if self._closing:
+                conn.close()
+            else:
+                self._free.append(conn)
+            asyncio.Task(self._wakeup(), loop=self._loop)
 
     @asyncio.coroutine
     def cursor(self, name=None, cursor_factory=None,
